import unittest

from transformers import T5TokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


@unittest.skip("T5Tokenizer's mapping is not well defined, not working")
class T5TokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = T5TokenizerFast
    pretrained_name = "t5-small"

    def setUp(self):
        super().setUp()



import unittest

from transformers import XGLMTokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class XGLMTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = XGLMTokenizerFast
    pretrained_name = "facebook/xglm-564M"

    def setUp(self):
        super().setUp()
        self.tokenizer.eos_token_id = None



import unittest

from transformers import BloomTokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


# @unittest.skip("GPTNeoXTokenizerFast is not available for testing")
class BloomTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = BloomTokenizerFast
    pretrained_name = "bigscience/bloom-560m"

    def setUp(self):
        super().setUp()



import copy
import logging
from abc import ABC
from functools import lru_cache
from typing import List

import torch

from transformers_cfg.recognizer import StringRecognizer, AcceptState
from transformers_cfg.parser import parse_ebnf
from transformers_cfg.trie import ByteTrie
from transformers_cfg.utf8_utils import PartialUTF8
from .vocab_struct import LEAF, TokenTrie
from transformers_cfg.mapping import get_mapping

logger = logging.getLogger(__name__)


class AbsTokenRecognizer(ABC):
    def __init__(self, grammar_str, tokenizer, start_rule_name="root", unicode=False):
        parsed_grammar = parse_ebnf(grammar_str)
        grammar_encoding = parsed_grammar.grammar_encoding
        self.start_rule_id = parsed_grammar.symbol_table.get(start_rule_name)
        self.byte_encoding = unicode

        if unicode and not tokenizer.__class__.__name__.lower().startswith(
            "gpt2"
        ):  # gpt2tokenizer or gpt2tokenizerfast
            raise ValueError(
                "Constrained decoding with unicode is only supported for GPT2 model. Support for other models is coming soon."
                "Or you can use the constraints with only ascii characters."
            )

        self.eos_token_id = tokenizer.eos_token_id
        self.token_trie = TokenTrie(tokenizer)
        self.tokenizer = tokenizer
        self.string_recognizer = StringRecognizer(grammar_encoding, self.start_rule_id)
        self.unicode_trie = ByteTrie.from_tokenizer(tokenizer, unicode=unicode)
        self.mapping = get_mapping(tokenizer, unicode=unicode)
        assert len(self.mapping) == len(
            self.token_trie
        ), f"{len(self.mapping)}, {len(self.token_trie)}"

    def _consume_token_id(
        self, token_id: int, accept_state: AcceptState
    ) -> AcceptState:
        if self.string_recognizer._must_stop(accept_state.stacks):
            if token_id == self.eos_token_id:
                return self.string_recognizer.get_termination_accept_state()
            else:
                raise ValueError(
                    f"All stacks are empty, so the only token accepted is EOS({self.eos_token_id}), but got {token_id}"
                )
        if token_id == self.eos_token_id:
            if self.string_recognizer._can_stop(accept_state.stacks):
                # if at least one of the stack is empty, we can stop
                # we clear all the stacks, meaning that we don't accept any token after EOS
                return self.string_recognizer.get_termination_accept_state()
            else:
                raise ValueError(
                    f"At least one of the stack should be empty when EOS is reached. However, "
                    f"the stacks are {accept_state.stacks}"
                )

        bytes_or_codepoints = self.mapping.map(token_id)
        accept_state = self.string_recognizer._consume_bytes(
            bytes_or_codepoints, accept_state
        )
        return accept_state

    def probe_token_id(self, token_id: int, accept_state: AcceptState) -> bool:
        stacks = accept_state.stacks
        if self.string_recognizer._must_stop(stacks):
            if token_id == self.eos_token_id:
                return True
            else:
                return False
        if token_id == self.eos_token_id:
            if self.string_recognizer._can_stop(stacks):
                # if at least one of the stack is empty, we can stop
                # we clear all the stacks, meaning that we don't accept any token after EOS
                return True
            else:
                return False
        # for code_point in self.mapping.map(token_id):
        #     stacks = self.grammar._consume_char_code_point(code_point, stacks)
        bytes_or_codepoints = self.mapping.map(token_id, verbose=False)
        new_acc_state = self.string_recognizer._consume_bytes(
            bytes_or_codepoints, accept_state, verbose=False
        )
        return len(new_acc_state.stacks) > 0

    def advance_token_ids(self, *args, **kwargs):
        """Process a list of tokens according to the grammar rules."""
        raise NotImplementedError

    def batch_filter_vocab(self, batch_accept_states, device) -> torch.Tensor:
        batch_acceptance = []
        for accept_state in batch_accept_states:
            batch_acceptance.append(self.filter_vocab(accept_state, device))
        return torch.stack(batch_acceptance)

    def filter_vocab(self, accept_state, device) -> torch.Tensor:
        if not accept_state.stacks:  # Check if stacks is empty
            # Handle the empty case: for example, return a tensor of False
            # The size of the tensor should match the size of your vocabulary
            vocab_size = len(self.mapping)
            logger.debug(f"Empty stack, sum of acceptance: {0}")
            # size of the vocab
            accepts = [False] * vocab_size
            accepts[self.eos_token_id] = True
            return torch.tensor(accepts, dtype=torch.bool, device=device)

        acceptance = self.get_token_acceptance(accept_state, device)

        return acceptance

    def get_token_acceptance(self, accept_state, device) -> torch.Tensor:
        acceptance_matrix = torch.cat(
            [
                self.get_token_acceptance_array_for_stack(
                    tuple(stack), accept_state.partial_utf8, device
                )
                for stack in accept_state.stacks
            ]
        )
        # Merge stacks: any True => True
        acceptance = acceptance_matrix.reshape(len(accept_state.stacks), -1).any(dim=0)
        return acceptance

    @lru_cache(maxsize=32768)
    def get_token_acceptance_array_for_stack(self, stack, partial_utf8, device):
        # stack = list(stack)  # needs to come in as a tuple for lru_cache
        assert isinstance(stack, tuple)
        stack = list(stack)

        if self.byte_encoding:

            accept_f = lambda x: self.string_recognizer._probe_bytes(
                x, [stack], partial_utf8=partial_utf8
            )
            token_acceptance = self.unicode_trie.get_token_acceptance(
                accept=accept_f, accept_eos=False, eos_token_id=self.eos_token_id
            )
        else:
            accepts = [False] * len(self.mapping)
            token_acceptance = check_token_acceptance_in_trie(
                self.token_trie.trie,
                [stack],
                self.string_recognizer,
                self.eos_token_id,
                accepts,
            )
        x = torch.tensor(token_acceptance, dtype=torch.bool, device=device)
        x_eos = self.validate_and_set_eos_acceptance(x)
        return x_eos

    def validate_and_set_eos_acceptance(self, acceptance: torch.Tensor) -> torch.Tensor:
        if torch.any(acceptance) == 0:
            acceptance[self.eos_token_id] = True
        else:
            if acceptance[self.eos_token_id]:
                raise ValueError()
            acceptance[self.eos_token_id] = False
        return acceptance


class IncrementalTokenRecognizer(AbsTokenRecognizer):
    def __init__(self, grammar_str, start_rule_name, tokenizer, unicode=False):
        super().__init__(grammar_str, tokenizer, start_rule_name, unicode)
        self.last_size = None

        # if self.last_size is not set (which would be the case when processing the first token).
        # In this case, do nothing.

    def advance_token_ids(self, input_ids, batch_accept_states, parse_start_index=None):

        if self.last_size is None:
            prefix_to_parse = [
                single_input_ids[parse_start_index:]
                if parse_start_index is not None
                else []
                for single_input_ids in input_ids
            ]

            # self.grammar_acceptor.accept_token_ids(prefix_to_parse, self.stacks)
            batch_accept_states = [
                self._consume_token_ids(prefix, accept_state)
                for prefix, accept_state in zip(prefix_to_parse, batch_accept_states)
            ]
            #  if the length of the current input IDs (input_ids[0]) is exactly one more than self.last_size.
            #  This is expected in a scenario where inputs are processed incrementally, one token at a time.
        elif len(input_ids[0]) == self.last_size + 1:
            batch_accept_states = [
                self._consume_token_id(single_input_ids[-1], accept_state)
                for single_input_ids, accept_state in zip(
                    input_ids, batch_accept_states
                )
            ]
            #  ensure that the input size is consistent with the expected incremental processing
            #  (i.e., one token at a time).
        else:
            # here we check if the input_ids are one token longer than the last time we processed
            # but we don't check if input_ids are actually valid.
            # Imagine a scenario where we generate 10 tokens, then we replace the 10 generated tokens with 10 new tokens.
            # In this case, the input_ids will be consistent with the last_size, but the input_ids are not valid.
            # However, should we really check if the input_ids are valid here?
            # If we do, then we need to reparse the whole input_ids at each call, which is not efficient.
            # Maybe we should just trust the user to provide valid input_ids?
            # The conclusion is that, we assume the input_ids are valid, and our generation will be correct.
            # If the input_ids are not valid, then the generation result will be wrong and we don't take responsibility for that.
            raise RuntimeError(
                "Input ID's length is inconsistent with the current state of "
                "the GrammarConstrainedLogitsProcessor. If you want to process "
                "another input sequence, please instantiate a new "
                "GrammarConstrainedLogitsProcessor."
            )
        self.last_size = len(input_ids[0])

        return batch_accept_states

    def _consume_token_ids(
        self, token_ids: List[int], accept_state: AcceptState = None, as_string=True
    ):
        if accept_state is None:
            accept_state = self.string_recognizer.get_initial_accept_state()
        if as_string:
            string = self.tokenizer.decode(token_ids)
            accept_state = self.string_recognizer._consume_string(string, accept_state)
        else:
            for i, token_id in enumerate(token_ids):
                accept_state = self._consume_token_id(token_id, accept_state)
                if len(accept_state.stacks) > 0:
                    cur_token_ids = token_ids[: i + 1]
                    logging.debug(f"{cur_token_ids} is accepted")
                    decoded_string = self.tokenizer.decode(cur_token_ids)
                    logging.debug(f"The decoded string is {decoded_string}")
        return accept_state


def check_token_acceptance_in_trie(trie, stacks, grammar, eos_token_id, accepts):

    for byte, next_trie in trie.items():
        if byte == LEAF:
            token_id = next_trie
            if token_id != eos_token_id:
                # if the stacks is not empty, it means we can still continue to parse
                # so we should accept the token
                accepts[token_id] = bool(stacks)
            continue

        new_stacks = []
        for stk in stacks:
            if not stk:
                continue

            next_element_offset = stk[-1]
            num_chars = grammar.grammar_encoding[next_element_offset]

            if not grammar.char_acceptance_at_element(next_element_offset).get(
                byte, False
            ):
                # if the current byte is not accepted by the current rule, we need to try next rule
                continue

            next_element_offset += num_chars + 1
            new_stack = stk[:-1]
            if grammar.grammar_encoding[next_element_offset]:
                new_stack.append(next_element_offset)
            new_stacks.extend(grammar.advance_stack(tuple(new_stack)))

        if new_stacks:
            check_token_acceptance_in_trie(
                next_trie, new_stacks, grammar, eos_token_id, accepts
            )

    return accepts


if __name__ == "__main__":
    from transformers import AutoTokenizer

    with open("examples/grammars/japanese.ebnf", "r") as file:
        input_text = file.read()
    parsed_grammar = parse_ebnf(input_text)
    parsed_grammar.print()

    tokenizer = AutoTokenizer.from_pretrained("gpt2")

    tokenRecognizer = IncrementalTokenRecognizer(
        grammar_str=input_text, start_rule_name="root", tokenizer=tokenizer
    )

    japanese = "トリーム"  # "こんにちは"
    token_ids = tokenizer.encode(japanese)
    # 13298, 12675, 12045, 254
    stacks = tokenRecognizer._consume_token_ids(
        token_ids, tokenRecognizer.string_recognizer.stacks, as_string=False
    )

    if stacks:
        print("The Japanese input is accepted")
    else:
        print("The Japanese input is not accepted")

    korean = "안녕하세요"
    token_ids = tokenizer.encode(korean)

    try:
        stacks = tokenRecognizer._consume_token_ids(
            token_ids, tokenRecognizer.string_recognizer.stacks, as_string=False
        )
        if stacks:
            print("The Korean input is accepted")
        else:
            print("The Korean input is not accepted")
    except ValueError as e:
        print("The Korean input is not accepted")



from typing import Dict, List

from transformers_cfg.utils import get_tokenizer_model_type, ints2bytes
from transformers import AutoTokenizer
import logging

log = logging.getLogger(__name__)


def get_mapping(tokenizer, unicode=False):
    log.debug(f"tokenizer type: {tokenizer.__class__.__name__}")
    log.debug(f"tokenizer model type: {get_tokenizer_model_type(tokenizer)}")
    if not unicode:
        if (
            "gpt2" in tokenizer.__class__.__name__.lower()
            or "bloom" in tokenizer.__class__.__name__.lower()
            or "pretrainedtokenizer" in tokenizer.__class__.__name__.lower()
            or "codegen" in tokenizer.__class__.__name__.lower()
            or "gptneox" in tokenizer.__class__.__name__.lower()
        ):
            return BBPEMapping(tokenizer)
        elif "t5" in tokenizer.__class__.__name__.lower():
            return BPEMapping(tokenizer)
        elif "llama" in tokenizer.__class__.__name__.lower():
            return LlamaBPEMapping(tokenizer)
        elif "xglm" in tokenizer.__class__.__name__.lower():
            return UniGramMapping(tokenizer)
        else:
            raise ValueError(f"Unknown tokenizer type: {tokenizer.__class__.__name__}")
    else:
        if "gpt2" in tokenizer.__class__.__name__.lower():
            return UnicodeBBPEMapping(tokenizer)
        else:
            raise NotImplementedError(
                f"Unicode mapping for {tokenizer.__class__.__name__}"
            )


class Mapping:
    def __init__(self, tokenizer):
        self.eos_token_id = tokenizer.eos_token_id
        self.bos_token_id = tokenizer.bos_token_id
        self.tokenizer = tokenizer
        self.special = tokenizer.all_special_ids

    def __len__(self):
        return len(self.tokenizer.get_vocab())

    def _map(self, token_id: int) -> str:
        # This is the case for BOS,
        if token_id in self.special:
            return ""
        # if token_id is tensor, convert it to int
        if hasattr(token_id, "item"):
            token_id = token_id.item()
        raw_token = self.tokenizer.convert_ids_to_tokens(token_id)
        return raw_token

    def map(self, token_id: int, verbose=False) -> bytes:
        token = self._map(token_id)
        if verbose:
            log.debug(f"token_id: {token_id}, token: {token}")
        return bytes(token, "utf-8")


class BBPEMapping(Mapping):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _map(self, token_id: int) -> str:
        raw_token = super()._map(token_id)
        if raw_token.startswith("Ġ"):
            raw_token = raw_token.replace("Ġ", " ")
        return raw_token


class UnicodeBBPEMapping(Mapping):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.intermediate_encoding = UnicodeBBPEMapping.get_intermediate_encoding(
            self.tokenizer
        )

    def _map(self, token_id: int, verbose=False) -> str:
        raw_token = super()._map(token_id)
        # if raw_token.startswith("Ġ"):
        #     raw_token = raw_token.replace("Ġ", " ")
        return raw_token

    def map(self, token_id: int, verbose=False) -> bytes:
        raw_token = self._map(token_id, verbose)
        if verbose:
            log.debug(f"token_id: {token_id}, raw_token: {raw_token}")
        return self.intermediate_encoding.token2bytes(raw_token)

    @staticmethod
    def get_intermediate_encoding(tokenizer):
        if "gpt2" in tokenizer.__class__.__name__.lower():
            return ByteEncoding(tokenizer)
        else:
            return None


class BPEMapping(Mapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.last_token_id = None

    def _map(self, token_id: int) -> str:
        raw_token = super()._map(token_id)

        # we need to check if the token is at the beginning of the sentence to remove the space
        # specific to BPE
        at_bos = False
        if self.last_token_id is not None and self.last_token_id == self.bos_token_id:
            at_bos = True
        self.last_token_id = token_id
        if raw_token.startswith("▁"):
            raw_token = raw_token.replace("▁", " ")
            if at_bos:
                # remove space at the beginning of the sentence
                raw_token = raw_token[1:]
        return raw_token


class LlamaBPEMapping(BPEMapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)

    def _map(self, token_id: int) -> str:
        raw_token = super()._map(token_id)
        # if the token is hex, token is a string like "<0x00>"
        # first 256 tokens are hex
        if raw_token.startswith("<0x"):
            hex_value = raw_token[4:-1]
            raw_token = chr(int(hex_value, 16))
        return raw_token


class WordPieceMapping(Mapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)

    def map(self, token_id: int) -> bytes:
        if token_id in self.special:
            return bytes()
        return bytes(
            self.tokenizer.decode([token_id], clean_up_tokenization_spaces=False),
            "utf-8",
        )


class UniGramMapping(Mapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)

    def map(self, token_id: int) -> bytes:
        if token_id in self.special:
            return bytes()
        return bytes(
            self.tokenizer.decode([token_id], clean_up_tokenization_spaces=False),
            "utf-8",
        )


class XGLMUniGramMapping(Mapping):
    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.bos_token_id = tokenizer.eos_token_id
        self.eos_token_id = None


class ByteEncoding:
    def __init__(self, tokenizer):
        # check if the tokenizer is fast, if so, convert it to slow
        if tokenizer.is_fast:
            tokenizer = AutoTokenizer.from_pretrained(
                tokenizer.name_or_path, use_fast=False
            )
        self.tokenizer = tokenizer
        self.byte2char: Dict[int, str] = tokenizer.byte_encoder
        self.char2byte: Dict[str, int] = tokenizer.byte_decoder
        # code point to byte
        self.cdp2byte: Dict[int, int] = {ord(c): b for c, b in self.char2byte.items()}
        self.byte2cdp: Dict[int, int] = {v: k for k, v in self.cdp2byte.items()}

    def map(self, byte: int) -> int:
        assert 0 <= byte < 256, f"byte: {byte} is not in the range [0, 256)"
        return ord(self.byte2char[byte])

    def token_ids2bytes(self, token_ids: List[int]) -> bytes:
        tokens: List[str] = self.tokenizer.convert_ids_to_tokens(token_ids)
        # for token id = BOS, the token should be empty string instead of <s>
        # TODO, this may cause issues because this means that special tokens like BOS can appear at any position
        tokens = [
            "" if token in self.tokenizer.all_special_ids else token for token in tokens
        ]
        bytes: List[List[int]] = [self.token2bytes(token) for token in tokens]
        # join the bytes
        return ints2bytes(sum(bytes, []))

    def token_id2bytes(self, token_id: int) -> bytes:
        token: str = self.tokenizer.convert_ids_to_tokens(token_id)
        return self.token2bytes(token)

    def token2bytes(self, token: str) -> bytes:
        # import pdb; pdb.set_trace()
        bytes_seq: List[int] = [self.char2byte[c] for c in token]
        return bytes(bytes_seq)


if __name__ == "__main__":
    gpt2_tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=True)

    gpt2_tokenizer.encode("榴莲")
    # [162, 99, 112, 164, 236, 110]

    mapping = get_mapping(gpt2_tokenizer)

    x = mapping.map(162)
    # b'\xef\xbf\xbd'
    x = mapping.map(99)

    ##################################

    gpt2_tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)

    token_ids = gpt2_tokenizer.encode("榴莲")

    print(token_ids)
    utf8_encoding = "榴莲".encode("utf-8")
    print("utf8_encoding: ", utf8_encoding)
    mapping = ByteEncoding(gpt2_tokenizer)
    print(mapping.token_ids2bytes(token_ids))
    # b'\xef\xbf\xbd'

    ###################################

    gpt2_tokenizer.encode("´")



import unittest

from transformers import PreTrainedTokenizer, AutoTokenizer

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class FalconTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = AutoTokenizer
    pretrained_name = "tiiuae/falcon-7b"

    def setUp(self):
        super().setUp()



import unittest

from transformers import GPT2TokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class GPT2TokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = GPT2TokenizerFast
    pretrained_name = "gpt2"

    def setUp(self):
        super().setUp()



import unittest
import warnings

from transformers import PreTrainedTokenizer
from transformers_cfg.token_grammar_recognizer import IncrementalTokenRecognizer

from transformers_cfg.recognizer import StringRecognizer

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.utils import pprint_token_ids


class TokenizerTesterMixin:
    tokenizer_class = None
    pretrained_name = None
    rust_tokenizer_class = None
    test_slow_tokenizer = True
    test_rust_tokenizer = True
    space_between_special_tokens = False
    from_pretrained_kwargs = None
    from_pretrained_filter = None
    from_pretrained_vocab_key = "vocab_file"
    test_seq2seq = True

    # set to True to test a sentencepiece tokenizer
    test_sentencepiece = False

    # set to True to ignore casing when testing a sentencepiece tokenizer
    # test_sentencepiece must also be set to True
    test_sentencepiece_ignore_case = False

    def setUp(self):
        self.tokenizer = self.get_tokenizer()

    def get_tokenizer(self, **kwargs) -> PreTrainedTokenizer:
        use_fast = kwargs.pop("use_fast", True)
        return self.tokenizer_class.from_pretrained(
            self.pretrained_name, use_fast=use_fast, **kwargs
        )

    def test_json_parsable(self):
        # Test that we can load a JSON object
        with open("examples/grammars/json.ebnf", "r") as file:
            input_text = file.read()
        JsontokenRecognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        valid_json = '{"foo": "bar", "baz": "bat"}'
        token_ids = self.tokenizer.encode(valid_json)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        acc_state = JsontokenRecognizer._consume_token_ids(token_ids, as_string=False)
        # the json object is complete, so the stacks should be empty
        self.assertTrue(
            acc_state.stacks == [] or acc_state.stacks == [[]],
            f"stacks: {acc_state.stacks}, not empty",
        )

    def test_balanced_parentheses(self):
        # Test that we can recognize a balanced parentheses
        with open("examples/grammars/balanced_parentheses.ebnf", "r") as file:
            input_text = file.read()
        recognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        balanced_parentheses = "((((((((()))))))))"
        token_ids = self.tokenizer.encode(balanced_parentheses)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        accept_state = recognizer._consume_token_ids(token_ids, as_string=False)
        # the json object is complete, so the stacks should be empty
        self.assertTrue(
            accept_state.stacks == [] or accept_state.stacks == [[]],
            f"stacks: {accept_state.stacks}, not empty",
        )

        # inbalanced_parentheses = "((((((((()))))))))))))"
        # token_ids = self.tokenizer.encode(inbalanced_parentheses)
        # pprint_token_ids(self.tokenizer, token_ids)
        #
        # # check if there is unk token
        # stacks = recognizer._consume_token_ids(
        #     token_ids, recognizer.grammar.stacks, as_string=False
        # )
        #
        # self.assertTrue(stacks != [] and stacks != [[]], f"stacks: {stacks}, empty")

    @unittest.skip("Not implemented")
    def test_emoji(self):
        """
        Test that we can accept emoji
        """

        with open("examples/grammars/emoji.ebnf", "r") as file:
            input_text = file.read()

        JsontokenRecognizer = IncrementalTokenRecognizer(
            grammar_str=input_text, start_rule_name="root", tokenizer=self.tokenizer
        )

        emoji = "😀😄😂"
        token_ids = self.tokenizer.encode(emoji)
        pprint_token_ids(self.tokenizer, token_ids)

        # check if there is unk token
        for token_id in token_ids:
            if token_id == self.tokenizer.unk_token_id:
                warnings.warn(
                    f"unk token found in input_token_ids: {token_ids}, skipping test"
                )
                return

        stacks = JsontokenRecognizer._consume_token_ids(
            token_ids, JsontokenRecognizer.string_recognizer.stacks, as_string=False
        )

        # parsed_grammar = parse_ebnf(input_text)
        #
        # start_rule_id = parsed_grammar.symbol_table["root"]
        #
        # recognizer = GrammarRecognizer(parsed_grammar.grammar_encoding, start_rule_id)
        #
        # self.assertTrue(recognizer._accept_string(emoji, recognizer.stacks))

    # def test_beam_search_low_memory(self):
    #     # Check that choosing 'low_memory' does not change the model output
    #     for model_class in self.all_generative_model_classes:
    #         if any(model_name in model_class.__name__.lower() for model_name in ["fsmt", "reformer"]):
    #             self.skipTest("Won't fix: old model with different cache format")
    #         if any(
    #             model_name in model_class.__name__.lower()
    #             for model_name in [
    #                 "bloom",
    #                 "ctrl",
    #                 "gptbigcode",
    #                 "transo_xl",
    #                 "xlnet",
    #                 "cpm",
    #             ]
    #         ):
    #             self.skipTest("May fix in the future: need model-specific fixes")
    #         config, input_ids, attention_mask, max_length = self._get_input_ids_and_config(batch_size=2)
    #         # batch_size=1 is ok, but batch_size>1 will cause non-identical output
    #
    #         config.use_cache = True
    #         config.is_decoder = True
    #
    #         # test output equality of low versus high memory
    #         model = model_class(config).to(torch_device).eval()
    #
    #         low_output = model.generate(input_ids, max_new_tokens=8, num_beams=5, early_stopping=True, low_memory=True)
    #
    #         high_output = model.generate(
    #             input_ids, max_new_tokens=8, num_beams=5, early_stopping=True, low_memory=False
    #         )
    #         self.assertListEqual(low_output.tolist(), high_output.tolist())



import unittest

from transformers import CodeGenTokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class CodeGenTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = CodeGenTokenizerFast
    pretrained_name = "microsoft/phi-1_5"

    def setUp(self):
        super().setUp()



import unittest

from transformers import GPTNeoXTokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class GPTNeoXTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = GPTNeoXTokenizerFast
    pretrained_name = "EleutherAI/pythia-160m-deduped"

    def setUp(self):
        super().setUp()



import unittest
from unittest import TestCase

from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers_cfg.token_grammar_recognizer import IncrementalTokenRecognizer
from transformers_cfg.generation import GrammarConstrainedLogitsProcessor

MODEL_IDS = [
    "hf-internal-testing/tiny-random-GPTJForCausalLM",
    # "hf-internal-testing/tiny-random-BloomForCausalLM",
    # "hf-internal-testing/tiny-random-PhiForCausalLM",
    "hf-internal-testing/tiny-random-gpt2",
    # "hf-internal-testing/tiny-random-BlenderbotForCausalLM",
]


class Test(TestCase):
    def test_grammar_constrained_decoding_greedy_w_number_grammar(self):
        # test greedy decoding with grammar constraints
        grammar_str = """
        root ::= [0-9]+
        """

        for model_id in MODEL_IDS:
            model = AutoModelForCausalLM.from_pretrained(model_id)
            tokenizer = AutoTokenizer.from_pretrained(model_id)
            tokenizer.pad_token = tokenizer.eos_token

            grammar = IncrementalTokenRecognizer(
                grammar_str, start_rule_name="root", tokenizer=tokenizer
            )
            grammar_processor = GrammarConstrainedLogitsProcessor(grammar)

            prefix = "This is a valid number:"

            input_ids = tokenizer(
                [prefix], add_special_tokens=False, return_tensors="pt", padding=True
            )["input_ids"]

            output = model.generate(
                input_ids,
                do_sample=False,
                pad_token_id=tokenizer.eos_token_id,
                eos_token_id=tokenizer.eos_token_id,
                num_beams=1,
                max_new_tokens=40,
                top_p=0.92,
                top_k=5,
                logits_processor=[grammar_processor],
                repetition_penalty=100.0,
                early_stopping=True,
            )

            # generations = tokenizer.batch_decode(output, skip_special_tokens=True)

            generations = tokenizer.batch_decode(
                output[:, input_ids.shape[1] :], skip_special_tokens=True
            )
            # assert generations[0].isdigit(), f"generations: {generations} is not a number"
            self.assertTrue(
                generations[0].isdigit(), f"generations: {generations} is not a number"
            )

    def test_grammar_constrained_decoding_greedy_w_balanced_parenthesis_grammar(self):
        # test greedy decoding with grammar constraints
        grammar_str = """
        root ::= "(" root ")" | ""
        """

        for model_id in MODEL_IDS:

            model = AutoModelForCausalLM.from_pretrained(model_id)
            tokenizer = AutoTokenizer.from_pretrained(model_id)
            tokenizer.pad_token = tokenizer.eos_token

            grammar = IncrementalTokenRecognizer(
                grammar_str, start_rule_name="root", tokenizer=tokenizer
            )
            grammar_processor = GrammarConstrainedLogitsProcessor(grammar)

            prefix = "This is a valid json:"

            input_ids = tokenizer(
                [prefix], add_special_tokens=False, return_tensors="pt", padding=True
            )["input_ids"]
            MAX_NEW_TOKENS = 20

            output = model.generate(
                input_ids,
                do_sample=True,
                pad_token_id=tokenizer.eos_token_id,
                eos_token_id=tokenizer.eos_token_id,
                num_beams=1,
                max_new_tokens=40,
                top_p=0.92,
                top_k=5,
                logits_processor=[grammar_processor],
                repetition_penalty=100.0,
                early_stopping=True,
            )

            generation: str = tokenizer.batch_decode(
                output[:, input_ids.shape[1] :], skip_special_tokens=True
            )[0]

            def check_parentheses(generation):
                stack = []
                for char in generation:
                    if char == "(":
                        stack.append(char)
                    elif char == ")":
                        if not stack:
                            return False
                        stack.pop()
                return not stack

            # assert check_parentheses(
            #     generation
            # ), f"generations: {generation} is not a balanced parenthesis"
            self.assertTrue(
                check_parentheses(generation),
                f"generations: {generation} is not a balanced parenthesis",
            )

            # raise correct error when the input is not a valid sequence
            tentative = "((())"
            input_ids = tokenizer(
                [tentative], add_special_tokens=True, return_tensors="pt", padding=True
            )["input_ids"]

            tokenRecognizer = IncrementalTokenRecognizer(
                grammar_str=grammar_str, start_rule_name="root", tokenizer=tokenizer
            )

            accept_state = tokenRecognizer._consume_token_ids(
                input_ids[0], as_string=False
            )
            # generations = tokenizer.batch_decode(output, skip_special_tokens=True)


#
#
# if __name__ == "__main__":
#     test_grammar_constrained_decoding_greedy_w_number_grammar()
#     test_grammar_constrained_decoding_greedy_w_balanced_parenthesis_grammar()



from .token_grammar_recognizer import IncrementalTokenRecognizer


# Old class name, kept for backward compatibility
class IncrementalGrammarConstraint(IncrementalTokenRecognizer):
    def __init__(self, *args, **kwargs):
        # import warnings
        # warnings.warn(
        #     "IncrementalGrammarConstraint is deprecated and will be removed in a future version. "
        #     "Please use IncrementalLLMGrammarRecognizer instead.",
        #     DeprecationWarning, stacklevel=2
        # )
        super().__init__(*args, **kwargs)



#################
# DATA STRUCTURES
#################

import logging
import re

logger = logging.getLogger(__name__)

LEAF = -1


class TokenTrie:
    def __init__(self, tokenizer):
        self.eos_token_id = tokenizer.eos_token_id
        self.tokens = []
        self.trie = {}
        self.load_tokens(tokenizer)

    def id2str(self, token_id):
        return self.tokens[token_id]

    def __len__(self):
        return len(self.tokens)

    def load_tokens(self, tokenizer):
        def replace_hex(match):
            hex_value = match.group(1)
            return chr(int(hex_value, 16))

        if "gpt2" in tokenizer.__class__.__name__.lower():
            special = tokenizer.additional_special_tokens_ids

            # Here, the decoder does a string replace on a bunch of sequences
            # like ' .' for '.'. This interferes with our assumptions, where a
            # token should always have exactly one representation.
            # Fortunately(?) text-generation-inference doesn't seem to run this
            # cleanup, so we get extraneous spaces. So, in order to generate
            # the right token set for TGI, we have to skip the space trimming.
            # See:
            # https://github.com/huggingface/transformers/blob/main/src/transformers/tokenization_utils_base.py#L3588-L3600
            def fmt_token(id):
                if id in special:
                    return None
                return bytes(
                    tokenizer.decode([id], clean_up_tokenization_spaces=False), "utf-8"
                )

        elif (
            "llama" in tokenizer.__class__.__name__.lower()
            or "t5" in tokenizer.__class__.__name__.lower()
        ):

            def fmt_token(id):
                token = tokenizer.convert_ids_to_tokens(id)
                token = re.sub(r"<0x([0-9a-fA-F]{2})>", replace_hex, token)
                token = token.replace("▁", " ")
                return bytes(token, "utf-8")

        else:
            logger.warning(
                "Warning: unrecognized tokenizer: using default token formatting"
            )

            def fmt_token(id):
                token = tokenizer.convert_ids_to_tokens(id)
                return bytes(token, "utf-8")

        # note: vocab_size doesn't work here because there are also
        # get_added_vocab() tokens
        self.tokens = [fmt_token(i) for i in range(len(tokenizer.get_vocab()))]
        for token_id, token_bytes in enumerate(self.tokens):
            if token_bytes is not None:
                self.insert_into_trie(self.trie, token_bytes, token_id)

    def insert_into_trie(self, trie, token_bytes, token_id):
        current = trie
        for byte in token_bytes:
            if byte not in current:
                current[byte] = {}
            current = current[byte]
        current[LEAF] = token_id


if __name__ == "__main__":
    from transformers import AutoTokenizer

    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    token_trie = TokenTrie(tokenizer)



import unittest

from transformers import LlamaTokenizerFast

from tests._tokenizer_common import TokenizerTesterMixin

import logging


class LlamaTokenizerTest(TokenizerTesterMixin, unittest.TestCase):

    tokenizer_class = LlamaTokenizerFast
    pretrained_name = "saibo/llama-1B"

    def setUp(self):
        super().setUp()
